//////////////////////////////////////////////////////////////////////////////
//	Module		: ClsIsa.cpp												//
//	Function	: Base class for ISA/VME/PC104 Smart A/D boards.			//
//	Author		: Jim Lamberson												//
//////////////////////////////////////////////////////////////////////////////

#include "ClsIsa.h"

//////////////////////////////////////////////////////////////////////////////
// Command/Data queue insertion/extraction functions.

void ClsIsa::SendByte( char val )
{
	*pCmdQ++ = val;			// Append byte to command queue.
}

///////////////////////////////////////////////////////////////////////////////
// Direct I/O functions.

char ClsIsa::InByteDirect(long Adrs)
{
	return InByte( Adrs );
}

void ClsIsa::OutByteDirect(long Adrs, char Value)
{
	OutByte( Adrs, Value );
}

bool ClsIsa::ActiveALARM()
{
	return ( InByte( MyAddress + 1 ) & STAT_ALARM ) != 0;
}

unsigned char ClsIsa::CmdStatusFlagsGet()
{
	return InByte( MyAddress + 1 ) & 0xF0;
}

////////////////////////////////////////////////////////////////////////////////////
// Wait for active handshake flag.

bool ClsIsa::HandshakeWait( char mask, char xormask, double maxtime )
{
	// Log the start time.
	double StartTime = SysTimer();

	// Wait for handshake flag or timeout, whichever occurs first.
	do
	{
		if ( ( ( InByte( MyAddress + 1 ) & mask ) ^ xormask ) != 0 )
			return true;
	}
	while ( ElapsedTime( StartTime ) < maxtime );

	// Max time elapsed, so we check handshake flag one final time in case this function
	// was interrupted for a long time.
	return ( ( ( InByte( MyAddress + 1 ) & mask ) ^ xormask ) != 0 );
}

///////////////////////////////////////////////////////////////////////////////
// Reset board and set FaultStuck flag if fault won't clear within one second.

#define MAX_RESET_RETRIES	3

void ClsIsa::CmdResetBoard()
{
	char	status;
	int		retry	= 0;

	// Repeat until the board's Fault flag is set or we reach max allowed retries ...
	while ( 1 )
	{
		// Invoke physical board hardware reset.
		OutByte( MyAddress + 1, 0 );

		// If the board's Fault flag is active, as it should be ...
		if ( InByte( MyAddress + 1 ) & STAT_FAULT )
		{
			// Wait for Fault to clear, or timeout, whichever occurs first.  If we timed out while waiting for Fault to clear ...
			if ( !HandshakeWait( STAT_FAULT, STAT_FAULT, 1.0 ) )
			{
				// Indicate problem with Fault: it won't switch to the inactive state.
				DeclareError( FAULT_LED );
			}
			else
			{
				// Cache the board's status byte.
				status = InByte( MyAddress + 1 );

				// Detect invalid DAV state.
				if ( ( status & STAT_DAV ) != 0 )
					DeclareError( FAULT_DAV );

				// Detect invalid CRMT state.
				if ( ( status & STAT_CRMT ) == 0 )
					DeclareError( FAULT_CRMT );
			}

			// Exit loop; we are done.
			break;
		}

		// Bump the retry counter.  If max allowed retries has been exceeded ...
		else if ( retry++ > MAX_RESET_RETRIES )
		{
			// Indicate problem with Fault: it won't switch to the active state.
			DeclareError( FAULT_LED );

			// Exit loop; we are done.
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Handshake any pending command string from CmdQ[] to board.

void ClsIsa::CmdClose()
{
	// If no CRMT fault previously detected ...
	if ( ( MyFaultFlags & FAULT_CRMT ) == 0 )
	{
		// Determine number of pending command bytes in CmdQ[], if any.
		short nCmdBytes = (short)( pCmdQ - CmdQ );

		// Handshake all pending command bytes to board.
		char *pCmdByte = CmdQ;
		for ( int i = 0; i < nCmdBytes; i++ )
		{
			// Wait for active CRMT or timeout.
			if ( !HandshakeWait( (const char)STAT_CRMT, 0, MAXHSTIME ) )
			{
				DeclareError( FAULT_CRMT );
				break;
			}

			// Send byte to Smart A/D board.
			OutByte( MyAddress, *pCmdByte++ );
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Fetch command response string from board into DatQ[].

void ClsIsa::HandshakeDatString( short nBytes = 0 )
{
	// If no DAV fault was previously detected ...
	if ( ( MyFaultFlags & FAULT_DAV ) == 0 )
	{
		// Repeat for each expected response byte ...
		for ( int i = 0; i < nBytes; i++ )
		{
			// Abort if handshake timed out.
			if ( !HandshakeWait( STAT_DAV, 0, MAXHSTIME ) )
			{
				DeclareError( FAULT_DAV );
				return;
			}

			// Fetch response byte and store in DatQ[].
			*pDatQ++ = InByte( MyAddress );
		}

		// Reset DatQ[] pointer so the first call to ReadByte() will return the proper value.
		pDatQ = DatQ;
	}
}

//////////////////////////////////////////////////////////////////////////////
// Send any pending command to board and fetch response string into DatQ[].

void ClsIsa::DeclareResponse( short nBytes )
{
	// Send any pending command string in CmdQ[] to board.
	CmdClose();

	// Fetch response string from board into DatQ[].
	HandshakeDatString( nBytes );

	// Reset CmdQ[] pointer in preparation for next command string.
	pCmdQ = CmdQ;
}
