//////////////////////////////////////////////////////////////////////////////
//	Module		: ClsNet.cpp												//
//	Function	: Base class for all Sensoray network Smart A/D boards.		//
//	Author		: Jim Lamberson												//
//	Created		: 07/01/1999												//
//////////////////////////////////////////////////////////////////////////////

#include "ClsNet.h"					// Class declaration.

/////////////////////////////////////////////////////////////////////////////
// Constants.
/////////////////////////////////////////////////////////////////////////////

#define FIRST_BASEPORT		0x2B0	// Base address of first 518 sub-board.
#define MAXPOLLTIME			30000	// Maximum time, in microseconds, to wait for sub-board DAV or CRMT.
#define MAX_WAIT_USEC		200000	// Maximum time, in microseconds, to wait for a response packet.
#define MAX_CMD_RETRIES		3		// Maximum number of times to retry a network transaction.

// 2509 command opcodes -----------------------

#define CMD_SETNETTOUT		0x01	// Set network communication timeout, in milliseconds.
#define CMD_READIO8			0x02	// Read byte from I/O port.
#define	CMD_LOOPBEG			0x03	// Begin 2509 command loop.
#define CMD_WRITEIO8		0x04	// Write byte to I/O port.
#define	CMD_LOOPEND			0x05	// End of 2509 command loop.
#define CMD_POLLT8			0x06	// Poll and mask byte port until non-zero or timeout.
#define	CMD_GETVERSION		0x07	// Return 2509 firmware version number.
#define CMD_POLLF8			0x08	// Poll and mask byte port until zero or timeout.
#define CMD_RESETCHK		0x09	// Test & clear 2509 Reset indicator.
#define CMD_SETPOLLTOUT		0x0A	// Set CMD_POLL/CMD_DELAY timeout, in microseconds.
#define CMD_DELAY			0x0C	// Delay until timeout.
#define CMD_EINT			0x10	// Enable interrupt.
#define CMD_DINT			0x12	// Disable interrupt.
#define CMD_ECHO8			0x13	// Echo byte data.

///////////////////////////////////////////////////////////////////////////////////
// Opcode and SDC tables are identical for all Ethernet SmartAD boards.

// Logical-to-Physical command opcode mapping table.
static const char CmdTbl[] = {
	(const char)0x00,	// CMD_RDCHAN	--	Get data from channel
	(const char)0x10,	// CMD_DFNTYP	--	Declare sensor type
	(const char)0x20,	// CMD_DFNLIM	--	Set alarm limits
	(const char)0x30,	// CMD_RDALRM	--	Get alarm flags for 8 channels
	(const char)0x40,	// CMD_RDTEMP	--	Get TB temperature(s)
	(const char)0x50,	// CMD_DFNOPN	--	Set open-sensor values
	(const char)0x70,	// CMD_TARE		--	Tare gage
	(const char)0x48,	// CMD_FILT50	--	Set 50Hz filter mode
	(const char)0x58,	// CMD_READ8	--	Get data from 8 channels
	(const char)0xFF,	//					 spare
	(const char)0xE0,	// CMD_CAL		--	Calibration function
	(const char)0xF0,	// CMD_DIAGNOSE	--	Diagnostic function
	(const char)0x80,	// CMD_RDSLOP	--	Get gage slope/intercept
	(const char)0x90,	// CMD_WTSLOP	--	Set gage slope/intercept
	(const char)0xB0,	// CMD_GZERO	--	Set gage zero
	(const char)0xD0,	// CMD_GSPAN	--	Set gage span
	(const char)0x60,	// CMD_STFILT	--	Set filter factor
	(const char)0x43,	// CMD_SLEEP	--	Switch to low-power standby
	(const char)0x42,	// CMD_WAKEUP	--	Wake from low-power standby
	(const char)0xC0	// CMD_WCOEFF	--	Set custom-R coefficients
};

// Logical-to-physical SDC mapping table.
static const char SdcTbl[] = {
	(const char)0xFF,	//  0 : 10V **
	(const char)0xFF,	//  1 : 10V
	(const char)0x00,	//  2 : 5V **
	(const char)0x15,	//  3 : 5V
	(const char)0x0E,	//  4 : 1.65V **
	(const char)0x16,	//  5 : 500mV
	(const char)0x17,	//  6 : 100mV
	(const char)0x0D,	//  7 : 80mV **
	(const char)0x01,	//  8 : E
	(const char)0x02,	//  9 : J **
	(const char)0x1B,	// 10 : J
	(const char)0x03,	// 11 : K **
	(const char)0x1C,	// 12 : K
	(const char)0x04,	// 13 : T **
	(const char)0x1D,	// 14 : T
	(const char)0x05,	// 15 : S **
	(const char)0x1E,	// 16 : S
	(const char)0x06,	// 17 : R **
	(const char)0x1F,	// 18 : R
	(const char)0x07,	// 19 : 385 **
	(const char)0x18,	// 20 : 385 platinum
	(const char)0x08,	// 21 : 392 **
	(const char)0x19,	// 22 : 392 platinum
	(const char)0x0B,	// 23 : 10K **
	(const char)0x1A,	// 24 : 10K
	(const char)0x11,	// 25 : 4-20 mA
	(const char)0x0F,	// 26 : Strain/pressure gage
	(const char)0x0A,	// 27 : 400 ohm range
	(const char)0x14,	// 28 : 4K ohm range
	(const char)0x20,	// 29 : 600K ohm range
	(const char)0x0C,	// 30 : Custom resistive sensor
	(const char)0x13,	// 31 : Disabled: remove from scan list
	(const char)0x22,	// 32 : N
	(const char)0x23,	// 33 : C
	(const char)0x24,	// 34 : B
	(const char)0xFF,	// 35 : 120 nickel RTD (STDbus only)
	(const char)0x28,	// 36 : 200 nickel RTD
	(const char)0x29,	// 37 : 1K nickel RTD
	(const char)0x2A,	// 38 : 385 platinum, HiRes (0.0125C)
	(const char)0x2B,	// 39 : 392 platinum, HiRes (0.0125C)
	(const char)0x25,	// 40 : 10K thermistor on fixed 600K ohm range
	(const char)0x2C,	// 41 : 10 ohm copper RTD
	(const char)0x14,	// 42 : 1K nickel RTD (4Kohm range, ext lin)
	(const char)0x14	// 43 : 1K platinum RTD (4Kohm range, ext lin)
};

//////////////////////////////////////////////////////////////////////////////
// Class constructor.

ClsNet::ClsNet( short ProductID, long IPAdrs, char NumBoards, short NumCh ) : ClsNonStd( ProductID, IPAdrs, NumCh, SdcTbl, CmdTbl, NumBoards )
{
	// Init various instance parameters.
	TransactionID	= 0;						// 2509 transaction identifier.

	// Init Model 518 sub-board parameters.
	ChansPerBoard	= NumCh / NumBoards;		// Sub-board channel count.
	GroupsPerBoard	= ChansPerBoard >> 3;		// Number of 8-channel groups per sub-board.
	for ( short b = 0; b < NumBoards; b++ )		// Sub-board's base address in 2509's PC/104 I/O space.
		BasePort[b] = FIRST_BASEPORT + ( b << 2 );
}

/////////////////////////////////////////////////////////////////////////////////////
//////////////////////////  SUB-BOARD SELECTION FUNCTIONS  //////////////////////////
/////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////
// Explicitly select a Model 518 sub-board.

void ClsNet::SelectBoard( short BoardNum )
{
	ActiveBasePort = BasePort[BoardNum];
}

//////////////////////////////////////////////////////////////////////////////////
// Select a Model 518 sub-board based on a channel number.
// Modify Chan so it references physical channel number on selected sub-board.

void ClsNet::SelectBoardByChan( short *Chan )
{
	ActiveBasePort = BasePort[ *Chan / ChansPerBoard ];
	*Chan &= ( ChansPerBoard - 1 );
}

////////////////////////////////////////////////////////////////////////////////
// Select a Model 518 sub-board based on an 8-channel group number.
// Modify Group so it references physical group number on selected sub-board.

void ClsNet::SelectBoardByGroup( short *Group )
{
	ActiveBasePort	= BasePort[ *Group / GroupsPerBoard ];
	*Group &= ( GroupsPerBoard - 1 );
}

////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////  COMMAND CONSTRUCTION FUNCTIONS  ///////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////
// Prepare to invoke a new 2509 command.

void ClsNet::CmdOpen()
{
	pSndBuf		= SndBuf + 2;	// Init NextWrite ptr to command string buffer (skip past TransactionID).
	nRespBytes	= 0;			// Indicate that no response data bytes are expected from server.

	// Set max time to wait for DAV/CRMT handshake before forcing error.
	MsgAppendCmd( CMD_SETPOLLTOUT, (short)MAXPOLLTIME );
}

////////////////////////////////////////////////////////////////////////////////////////////
// Append command string to transmit packet buffer:  Handshake cmd byte to SmartAD.
// Assumed: CMD_SETPOLLTOUT previously sent to 2509 or prepended to current command string.

void ClsNet::SendByte( char Value )
{
	// Append SioNet command sequence to transmit command packet buffer:
	MsgAppendCmd( CMD_POLLT8, ActiveBasePort + 1, (char)STAT_CRMT );	// Wait for CRMT or timeout.
	MsgAppendCmd( CMD_WRITEIO8, ActiveBasePort, Value );				// Write byte to sub-board command register.
}

////////////////////////////////////////////////////////////////////////////////////////////
// Append the specified number of HandshakedRead requests to transmit packet buffer.
// Assumed: CMD_SETPOLLTOUT previously sent to 2509 or prepended to current command string.

void ClsNet::DeclareResponse( short nBytes )
{
	switch ( nBytes )
	{
	case 0:
		return;

	case 1:
		// Append one handshake command sequence to SndBuf[].
		MsgAppendCmd( CMD_POLLT8, ActiveBasePort + 1, (char)STAT_DAV );	// Wait for DAV or timeout, whichever occurs first.
		MsgAppendREADIO8( ActiveBasePort );								// Read byte from sub-board data register.
		break;

	default:
		// Append iterative handshake command sequence to SndBuf[].
		MsgAppendCmd( CMD_LOOPBEG, nBytes );							// Begin 2509 command loop.
		MsgAppendCmd( CMD_POLLT8, ActiveBasePort + 1, (char)STAT_DAV );	//   Wait for DAV or timeout, whichever occurs first.
		MsgAppendCmd( CMD_READIO8, ActiveBasePort );					//   Read byte from sub-board data register.
		MsgAppendCmd( CMD_LOOPEND );									// End 2509 command loop.

		// Bump number of expected response bytes.
		nRespBytes += nBytes;
		break;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////  PACKET TRANSMIT/RECEIVE FUNCTIONS  ////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////
// Invoke execution of pending command string and fetch command response string.

void ClsNet::CmdClose()
{
	// Abort if transaction faults were previously detected.
	if ( MyFaultFlags & ( FAULT_SEND | FAULT_RECEIVE ) )
		return;

	// Attempt transactions until successful or max retries exceeded.
	for ( int retries = 0; retries < MAX_CMD_RETRIES; retries++ )
	{
		// Bump transaction ID and prepend it to header.  This is done so that all transactions, including
		// retries, will have a unique ID that can be used to match a response packet to its command packet.
		TransactionID++;
		SndBuf[0] = LOBYTE( TransactionID );
		SndBuf[1] = HIBYTE( TransactionID );

		// Attempt to send command packet in SndBuf[] to 2509 for execution, and if successful ...
		if ( PacketSend( SndBuf, (long)(pSndBuf - SndBuf), MAX_WAIT_USEC ) )
		{
			// If no reply is expected then the transaction is finished.
			if ( nRespBytes == 0 )
				return;
			
			// Attempt to receive the matching response packet into DatQ[]; exit if successful.
			if ( MsgRcv() )
				return;
		}
		else
		{
			// Transmit failed, so register the error and terminate the transaction.  This is probably caused
			// by a buffer allocation problem in the socket.
			DeclareError( FAULT_SEND );
			return;
		}
	}

	// Too many transaction retries; the Smart A/D isn't responding or there is a network problem.
	DeclareError( FAULT_RECEIVE );
}

/////////////////////////////////////////////////////////////////////////////
// Wait for received packet, then copy packet data to DatQ[].
// Returns true if packet was successfully received.

bool ClsNet::MsgRcv()
{
	// Repeat until either (1) a valid response packet is received, or (2) the socket's receive queue is empty and
	// we have waited the worst-case time required for a response packet to arrive.
	for ( ; ; )
	{
		// Reset read pointer to the beginning of receive packet buffer.  This must be repeated each pass through the
		// loop because ReadWordRev() modifies the pointer.
		pDatQ = DatQ;

		// Attempt to fetch the next pending received packet into DatQ[].  The socket will block this thread until
		// either a packet becomes available or the wait times out.
		short rsize = PacketReceive( DatQ, sizeof(DatQ), MAX_WAIT_USEC );

		// If no packets were pending, exit with indication that packet was not received.
		if ( rsize == 0 )
			return false;		// Packet receive failed.

		// Validate the received packet.  If not valid, drop the packet and loop back for the next one ------------------

		// Compare actual and expected packet sizes (adjusted for 2-byte TransactionID header).  If size is correct ...
		if ( rsize == ( nRespBytes + 2 ) )
		{
			// Validate received TransactionID against that of the transmitted command packet and exit if they match.  If IDs
			// don't match, we will drop this packet and loop back for the next one.
			if ( TransactionID == ReadWordRev() )
				return true;		// Packet successfully received.
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////  SOFT RESET & STATUS FUNCTIONS  //////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////
// Return true if any sub-board is sounding an alarm.

bool ClsNet::ActiveALARM()
{
	return ( (CmdStatusFlagsGet() & STAT_ALARM) != 0 );
}

//////////////////////////////////////////////////////////
// Fetch and OR together all sub-board status registers.

unsigned char ClsNet::CmdStatusFlagsGet()
{
	// Fetch status register images of all sub-boards.
	FetchStatusRegs();

	// Combine images by or'ing.
	char Status = 0;
	for ( int b = 0; b < MyNumBoards; b++ )
		Status |= ReadByte();

	// Return combined images.
	return Status;
}

////////////////////////////////////////////////////////////
// Fetch and AND together all sub-board status registers.

char ClsNet::AndStatusRegs()
{
	// Fetch status register images of all sub-boards.
	FetchStatusRegs();

	// Combine images by anding.
	char Status = (char)0xFF;
	for ( int b = 0; b < MyNumBoards; b++ )
		Status &= ReadByte();

	// Return combined images.
	return Status;
}

///////////////////////////////////////////////////
// Get images of all sub-board status registers.

void ClsNet::FetchStatusRegs()
{
	// Fetch status register images of all sub-boards.
	CmdOpen();
	for ( int b = 0; b < MyNumBoards; b++ )
		MsgAppendREADIO8( BasePort[b] + 1 );
	CmdClose();
}

///////////////////////////////////////////////////////////////////////////////
// Reset board and validate proper functioning of Status register.
///////////////////////////////////////////////////////////////////////////////

#define RESET_RETRIES	2		// Number of times to retry (in case 2509 is undergoing reset when we first try).

void ClsNet::CmdResetBoard()
{
	long errflags;

	// Attempt to reset Smart A/D, exit if successful.
	if ( ( errflags = TryReset() ) == 0 )
		return;

	// If we have not exhausted all retries ...
	for ( int i = 1; i < RESET_RETRIES; i++ )
	{
		// Delay one second to allow everything to reach equilibrium.
		double StartTime = SysTimer();
		while ( ElapsedTime(StartTime) < 1.0 );

		// Attempt to reset Smart A/D and exit if successful.
		if ( ( errflags = TryReset() ) == 0 )
			return;
	}

	// All retries have be exhausted, so log the problem and bail out.
	DeclareError( errflags );
}

long ClsNet::TryReset()
{
	// Indicate no pending board faults.
	long errflags = 0;
	MyFaultFlags &= ~( FAULT_DAV | FAULT_CRMT | FAULT_LED | FAULT_SEND | FAULT_RECEIVE );

	// On all model 518 sub-boards, invoke soft reset and then immediately fetch status register image.
	CmdOpen();
	for ( char b = 0; b < MyNumBoards; b++ )
	{
		MsgAppendCmd( CMD_WRITEIO8, BasePort[b] + 1, 0 );	// Invoke soft reset.
		MsgAppendREADIO8( BasePort[b] + 1 );				// Return 518 status register.
	}
	CmdClose();

	// If SoftReset Ethernet transaction was successful ...
	if ( ( MyFaultFlags & ( FAULT_SEND | FAULT_RECEIVE ) ) == 0 )
	{
		// Combine all model 518 status register Fault flags.
		char Status = STAT_FAULT;
		for ( b = 0; b < MyNumBoards; b++ )
			Status &= ReadByte();

		// If one or more sub-board Fault LEDs never turned on ...
		if ( Status == 0 )
			errflags |= FAULT_LED;		// Indicate problem with Fault status.
		else
		{
			// Repeat until all sub-board Fault LEDs turn off ...
			for ( double StartTime = SysTimer(); CmdStatusFlagsGet() & STAT_FAULT != 0; )
			{
				// If more than one second has elapsed ...
				if ( ElapsedTime(StartTime) > 1.0 )
				{
					// One or more Fault LEDs won't turn off, so indicate problem with Fault status and break out of test.
					errflags |= FAULT_LED;
					break;
				}
			}
		}

		// Detect any invalid handshake flag states.
		if ( CmdStatusFlagsGet() & STAT_DAV   )
			errflags |= FAULT_DAV;

		if ( !( AndStatusRegs() & STAT_CRMT ) )
			errflags |= FAULT_CRMT;
	}

	// Return any detected fault flags.
	return errflags;
}

/////////////////////////////////////////////////////////////////////////////////
///////////////////////  MESSAGE CONSTRUCTION UTILITIES  ////////////////////////
/////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////
// Append word value to command message.

void ClsNet::MsgAppendWord( short value )
{
	*pSndBuf++ = LOBYTE( value );
	*pSndBuf++ = HIBYTE( value );
}

////////////////////////////////////////////////////////////////
// OVERLOADED FUNCTION: Append command to command message.

void ClsNet::MsgAppendCmd( char OpCode )
{
	*pSndBuf++ = OpCode;
}

void ClsNet::MsgAppendCmd( char OpCode, char bVal )
{
	*pSndBuf++ = OpCode;
	*pSndBuf++ = bVal;
}

void ClsNet::MsgAppendCmd( char OpCode, short wVal )
{
	*pSndBuf++ = OpCode;
	*pSndBuf++ = LOBYTE( wVal );
	*pSndBuf++ = HIBYTE( wVal );
}

void ClsNet::MsgAppendCmd( char OpCode, short wVal, char bVal )
{
	*pSndBuf++ = OpCode;
	*pSndBuf++ = LOBYTE( wVal );
	*pSndBuf++ = HIBYTE( wVal );
	*pSndBuf++ = bVal;
}

///////////////////////////////////////////////
// Append I/O READ to command message.

void ClsNet::MsgAppendREADIO8( short PortAdrs )
{
	MsgAppendCmd( CMD_READIO8, PortAdrs );
	nRespBytes++;
}

///////////////////////////////////////////////////////////
// DIAGNOSTIC: Output byte to a local 2509 I/O port.
void ClsNet::OutByteDirect( long Adrs, char Value )
{
	CmdOpen();
	MsgAppendCmd( CMD_WRITEIO8, (short)Adrs, Value );
	CmdClose();
}

/////////////////////////////////////////////////////////
// DIAGNOSTIC: Input byte from a local 2509 I/O port.

char ClsNet::InByteDirect( long Adrs )
{
	CmdOpen();
	MsgAppendREADIO8( (short)Adrs );
	CmdClose();
	return ReadByte();
}

/////////////////////////////////////////////////////////////////////////////////////////
// Fetch product ID from Smart A/D board.  Since network Smart A/D products can not be
// physically queried for a product identifier, the product ID that was declared upon
// object instantiation is used.

short ClsNet::ProductIDGet()
{
	return MyProductID;
}

///////////////////////////////////////////////////////////////////////////
// Return 2509 firmware version number.

short ClsNet::NetVersionGet()
{
	CmdOpen();
	MsgAppendCmd( CMD_GETVERSION );			// Append GetVersion command to string.
	nRespBytes += 2;						// Indicate number of expected response bytes.
	CmdClose();
	return ReadWordRev();
}

///////////////////////////////////////////////////////////////////////////
// Check and reset 2509 Reset status flag.
// Return True if 2509 unit was reset since last NetResetChk() call.

bool ClsNet::NetResetChk()
{
	// Fetch the board's reset status.
	CmdOpen();
	MsgAppendCmd( CMD_RESETCHK );			// Append ResetChk command to string.
	nRespBytes++;							// Indicate number of expected response bytes.
	CmdClose();

	// Set up return value.
	bool rtnval = ( ReadByte() != 0 );

	// If unit was reset, init all board properties.
	if ( rtnval )
		ResetBoardProperties();

	return rtnval;
}

///////////////////////////////////////////////////////////////////////////
// Set 2509 network inactivity timeout, in milliseconds.

void ClsNet::NetTimeoutSet( long msec )
{
	// Limit time interval to max legal value.
	if ( msec > 0x00FFFFFF )
		msec = 0x00FFFFFF;

	// Enable new timeout interval.
	CmdOpen();
	MsgAppendCmd( CMD_SETNETTOUT, (short)msec, (char)( msec >> 16 ) );
	CmdClose();

	MyNetTimeout = msec;
}

long ClsNet::NetTimeoutGet()
{
	return MyNetTimeout;
}

