//////////////////////////////////////////////////////////////////////////////////
//	Module	 : ClsNetWin.h														//
//  Function : Windows interface template class for Ethernet Smart A/D boards.	//
//	Author	 : Jim Lamberson													//
//////////////////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <winsock.h>
#include <time.h>

//////////////////////////////////////////////////////////////////////////////
// Class declaration.
//////////////////////////////////////////////////////////////////////////////

template <class CBase> class ClsNetWin : public CBase {
public:
	ClsNetWin( long IPAdrs, long NicAdrs = INADDR_ANY );
	~ClsNetWin();
protected:
	short	PacketReceive	( char *pBuf, short BufLen, long usec )	;
	bool	PacketSend		( char *pBuf, long SndLen, long usec )	;
	void	SysSleep		( long msec )							{ Sleep(msec); }
	double	SysTimer		()										{ return (DOUBLE)clock() / CLOCKS_PER_SEC; }
private:
	// Local storage:
	bool		SocketIsOpen;		// WSAStartup() was successful.
	bool		MyIsRegistered;		// Socket created and bound to local IP adrs.
	SOCKADDR_IN ServerAddr;			// IP address of Net Smart A/D server.
	SOCKET		hSocket;			// Handle to socket.
};

// Manifest constants:
#define VERSION_REQD		0x0101			// We require Windows Socket v1.01 minimum.
#define MIN_SOCKETS_REQD	4				// Number of simultaneous sockets, minimum.

//////////////////////////////////////////////////////////////////////////////////////////////
// Constructor: create a socket and bind it to local address.  In order to bind to
// a multi-homed host's non-default NIC, NicAdrs must be set to the target NIC's address.
// If NicAdrs is not specified or is set to zero then the host's default NIC will be used.

template <class CBase>
ClsNetWin< CBase >::ClsNetWin( long IPAdrs, long NicAdrs ) : CBase( IPAdrs )
{
	WSAData	WSA_data;

	MyIsRegistered = SocketIsOpen = FALSE;

	// Start WinSock and fill the WSA_data structure with the attributes of the WinSock driver.
	if ( !( SocketIsOpen = ( WSAStartup(VERSION_REQD, &WSA_data) == 0 ) ) )
	{
		DeclareError( FAULT_OPEN );
		return;
	}

	// Verify that WinSock version number is at least the required minimum version number.
	if ( ( ( LOBYTE(WSA_data.wVersion) << 8 ) | HIBYTE(WSA_data.wVersion) ) < VERSION_REQD )
	{
		DeclareError( FAULT_OPEN );
		return;
	}

	// Verify that WinSock supports the required number of connections.
	if ( WSA_data.iMaxSockets < MIN_SOCKETS_REQD )
	{
		DeclareError( FAULT_OPEN );
		return;
	}

	// Create a socket to talk to the Ethernet Smart A/D.
	if ( ( hSocket = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP) ) == INVALID_SOCKET )
	{
		DeclareError( FAULT_OPEN );
		return;
	}

	// Bind the socket to a NIC.
	SOCKADDR_IN	ClientAddr;							// Allocate socket's address structure and populate it:
	ClientAddr.sin_family = AF_INET;				//   Address type is IP.
	ClientAddr.sin_addr.s_addr = htonl( NicAdrs );	//   Use the NIC at this address, or default NIC if address is not specified.
	ClientAddr.sin_port = 0;						//   Let the socket choose an unused local port.
	if ( bind( hSocket, (LPSOCKADDR)&ClientAddr, sizeof(ClientAddr) ) )
	{
		DeclareError( FAULT_REG );
		return;
	}

	// Initialize remote Net server address structure.
	ServerAddr.sin_family = AF_INET;
	ServerAddr.sin_addr.s_addr = htonl( IPAdrs );
	ServerAddr.sin_port = 0;

	// Indicate Smart A/D is successfully registered.
	MyIsRegistered = TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// Destructor: shut down socket and clean up.

template <class CBase>
ClsNetWin< CBase >::~ClsNetWin()
{
	if ( SocketIsOpen )
		WSACleanup();
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
// Transmit packet in SndBuf[] of size SndLen, return TRUE upon successful transmit.
// Block calling thread for up to the specified microseconds for an available transmit buffer.

template <class CBase>
bool ClsNetWin< CBase >::PacketSend( char *SndBuf, long SndLen, long usec )
{
	fd_set	EnabledFDSet;				// List of sockets to check.
	timeval	TimeValue = { 0, usec };	// Time to wait in select() before timeout.

	// Init list of sockets to be tested for transmit readiness.
	EnabledFDSet.fd_array[0] = hSocket;
	EnabledFDSet.fd_array[1] = INVALID_SOCKET;
	EnabledFDSet.fd_count = 1;

	// Transmit command packet if a transmit buffer becomes available within max allowed time.
	if ( select( 0, 0, &EnabledFDSet, 0, &TimeValue ) == 1 )
		return ( (long)sendto( hSocket, (const char *)SndBuf, SndLen, 0, (LPSOCKADDR)&ServerAddr, sizeof(ServerAddr) ) == SndLen );
	else
		return false;
}

/////////////////////////////////////////////////////////////////////////////////////
// Receive packet into Buf[] of max size BufLen, return received byte count.
// Block for up to the specified microseconds for a received packet.

template <class CBase>
short ClsNetWin< CBase >::PacketReceive( char *pBuf, short BufLen, long usec )
{
	fd_set	EnabledFDSet;				// List of sockets to check.
	timeval	TimeValue = { 0, usec };	// Time to wait in select() before timeout.

	// Init list of sockets to be tested for receive readiness.
	EnabledFDSet.fd_array[0] = hSocket;
	EnabledFDSet.fd_array[1] = INVALID_SOCKET;
	EnabledFDSet.fd_count = 1;

	// If packet is received within the max allowed time, fetch it and return its byte count, otherwise return zero.
	if ( select( 0, &EnabledFDSet, 0, 0, &TimeValue ) == 1 )
		return (short)recvfrom( hSocket, (char *)pBuf, (int)BufLen, 0, 0, 0 );
	else
		return 0;
}
