//////////////////////////////////////////////////////////////////////////////
//	Module		: ClsPci.cpp												//
//	Function	: Base class for Sensoray PCI Smart A/D boards.				//
//	Author		: Jim Lamberson												//
//	Created		: 04/14/1999												//
//////////////////////////////////////////////////////////////////////////////

#include "ClsPci.h"						// Class declaration.

// I/O constants for PCI products -----------------------------------------
#define CMDBUF				0x000000F0L	// Address offset of cmd buffer in Smart A/D local RAM window.
#define CNTRL_IMAGE			0x00684042L	// Default image of Misc. control register.
										// Address offsets from Plug-and-Play base addresses:
#define MEM_HANDSHAKE		0x000000EBL	//   Handshake control port: writes negate CRMT, reads negate DAV.
#define CNTRL_ADRS			0x00000050L	//   Misc. control register.

#define CNTRL_MSK_EECS		0x02000000L
#define CNTRL_MSK_RESET		0x40000000L

// Status register macros -------------------------------------------------

#define HOST_STATUS			0x0000001CL	// Host status port accessed via Local I/O window.

/////////////////////////////////////////////////////////////////////////////////////
// Opcode and SDC mapping tables are identical for all PCI products.

// Logical-to-Physical command opcode mapping table.
static const char CmdTbl[] = {
	(const char)0x00,	// CMD_RDCHAN	--	Get data from channel
	(const char)0x10,	// CMD_DFNTYP	--	Declare sensor type
	(const char)0x20,	// CMD_DFNLIM	--	Set alarm limits
	(const char)0x30,	// CMD_RDALRM	--	Get alarm flags for 8 channels
	(const char)0x40,	// CMD_RDTEMP	--	Get TB temperature(s)
	(const char)0x50,	// CMD_DFNOPN	--	Set open-sensor values
	(const char)0x70,	// CMD_TARE		--	Tare gage
	(const char)0x48,	// CMD_FILT50	--	Set 50Hz filter mode
	(const char)0x58,	// CMD_READ8	--	Get data from 8 channels
	(const char)0xFF,	//					 spare
	(const char)0xE0,	// CMD_CAL		--	Calibration function
	(const char)0xF0,	// CMD_DIAGNOSE	--	Diagnostic function
	(const char)0x80,	// CMD_RDSLOP	--	Get gage slope/intercept
	(const char)0x90,	// CMD_WTSLOP	--	Set gage slope/intercept
	(const char)0xB0,	// CMD_GZERO	--	Set gage zero
	(const char)0xD0,	// CMD_GSPAN	--	Set gage span
	(const char)0x60,	// CMD_STFILT	--	Set filter factor
	(const char)0x43,	// CMD_SLEEP	--	Switch to low-power standby
	(const char)0x42,	// CMD_WAKEUP	--	Wake from low-power standby
	(const char)0xC0	// CMD_WCOEFF	--	Set custom-R coefficients
};

// Logical-to-physical SDC mapping table.
static const char SdcTbl[] = {
	(const char)0x00,	//  0 : 10V **
	(const char)0x21,	//  1 : 10V
	(const char)0x00,	//  2 : 5V **
	(const char)0x15,	//  3 : 5V
	(const char)0x0E,	//  4 : 1.65V **
	(const char)0x16,	//  5 : 500mV
	(const char)0x17,	//  6 : 100mV
	(const char)0x0D,	//  7 : 80mV **
	(const char)0x01,	//  8 : E
	(const char)0x02,	//  9 : J **
	(const char)0x1B,	// 10 : J
	(const char)0x03,	// 11 : K **
	(const char)0x1C,	// 12 : K
	(const char)0x04,	// 13 : T **
	(const char)0x1D,	// 14 : T
	(const char)0x05,	// 15 : S **
	(const char)0x1E,	// 16 : S
	(const char)0x06,	// 17 : R **
	(const char)0x1F,	// 18 : R
	(const char)0x07,	// 19 : 385 **
	(const char)0x18,	// 20 : 385 platinum
	(const char)0x08,	// 21 : 392 **
	(const char)0x19,	// 22 : 392 platinum
	(const char)0x0B,	// 23 : 10K **
	(const char)0x1A,	// 24 : 10K
	(const char)0x11,	// 25 : 4-20 mA
	(const char)0x0F,	// 26 : Strain/pressure gage
	(const char)0x0A,	// 27 : 400 ohm range
	(const char)0x14,	// 28 : 4K ohm range
	(const char)0x20,	// 29 : 600K ohm range
	(const char)0x0C,	// 30 : Custom resistive sensor
	(const char)0x13,	// 31 : Disabled: remove from scan list
	(const char)0x22,	// 32 : N
	(const char)0x23,	// 33 : C
	(const char)0x24,	// 34 : B
	(const char)0xFF,	// 35 : 120 nickel RTD (STDbus only)
	(const char)0x28,	// 36 : 200 nickel RTD
	(const char)0x29,	// 37 : 1K nickel RTD
	(const char)0x2A,	// 38 : 385 platinum, HiRes (0.0125C)
	(const char)0x2B,	// 39 : 392 platinum, HiRes (0.0125C)
	(const char)0x25,	// 40 : 10K thermistor on fixed 600K ohm range
	(const char)0x2C,	// 41 : 10 ohm copper RTD
	(const char)0x14,	// 42 : 1K nickel RTD (4Kohm range, ext lin)
	(const char)0x14	// 43 : 1K platinum RTD (4Kohm range, ext lin)
};

///////////////////////////////////////////////////////////////////////////////
// Constructor.

ClsPci::ClsPci( short ProductID, long Adrs, short NumCh ) : ClsNonStd( ProductID, Adrs, NumCh, SdcTbl, CmdTbl )
{
	// Base class does all initialization.
}

//////////////////////////////////////////////////////////////////////////////
// Command/Data queue insertion/extraction functions.

void ClsPci::SendByte( char val )
{
	*pCmdQ++ = val;			// Append byte to command queue.
}





#define DELAY 					for ( int delay = 0; delay < 10; delay++ ) InPortByte( CfgBase );

#define IN_PORT_BYTE(ADRS)		InPortByte( ADRS ); DELAY

#define OUT_PORT_BYTE(ADRS,VAL)	OutPortByte( (ADRS), (VAL) ); DELAY




///////////////////////////////////////////////////////////////////////////////
// Reset board and set FaultStuck flag if fault won't clear within one second.

#define MAX_RESET_RETRIES	3

void ClsPci::CmdResetBoard()
{
	int		retry	= 0;
	char	status	= 0;

	// Repeat until the board's Fault flag is set or we reach max allowed retries ...
	while ( 1 )
	{
		// Assert & negate Local Reset.
		OutPortDword( CfgBase + CNTRL_ADRS, CNTRL_IMAGE | CNTRL_MSK_RESET );
		OutPortDword( CfgBase + CNTRL_ADRS, CNTRL_IMAGE );

		// Assert & negate EECS output to init EE arbiter state.
		OutPortDword( CfgBase + CNTRL_ADRS, CNTRL_IMAGE | CNTRL_MSK_EECS );
		OutPortDword( CfgBase + CNTRL_ADRS, CNTRL_IMAGE );

		// If the board's Fault flag is active, as it should be ...
		char status = IN_PORT_BYTE( IoBase + HOST_STATUS );
		if ( status & STAT_FAULT )
		{
			// Wait for Fault to clear, or timeout, whichever occurs first.  If we timed out while waiting for Fault to clear ...
			if ( !HandshakeWait( STAT_FAULT, STAT_FAULT, 1.0 ) )
			{
				// Indicate problem with Fault: it won't switch to the inactive state.
				DeclareError( FAULT_LED );
			}
			else
			{
				// Cache the board's status byte.
				status = IN_PORT_BYTE( IoBase + HOST_STATUS );

				// Detect invalid DAV state.
				if ( ( status & STAT_DAV ) != 0 )
					DeclareError( FAULT_DAV );

				// Detect invalid CRMT state.
				if ( ( status & STAT_CRMT ) == 0 )
					DeclareError( FAULT_CRMT );
			}

			// Exit loop; we are done.
			break;
		}

		// Bump the retry counter.  If max allowed retries has been exceeded ...
		else if ( retry++ > MAX_RESET_RETRIES )
		{
			// Indicate problem with Fault: it won't switch to the active state.
			DeclareError( FAULT_LED );

			// Exit loop; we are done.
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
// Return TRUE if alarm flag is active.

bool ClsPci::ActiveALARM()
{
	char status = IN_PORT_BYTE( IoBase + HOST_STATUS )
	return ( ( status & STAT_ALARM ) != 0 );
}

///////////////////////////////////////////////////////////////////////////
// Return status flags:  (0x00000,0,0,0,REG,CRMT,DAV,ALARM,FAULT,0x0).

unsigned char ClsPci::CmdStatusFlagsGet()
{
	char status = IN_PORT_BYTE( IoBase + HOST_STATUS );
	return (unsigned char)status;
}

////////////////////////////////////////////////////////////////////////////////////
// Wait for active handshake flag.

bool ClsPci::HandshakeWait( char mask, char xormask, double maxtime )
{
	char status;

	// Log the start time.
	double StartTime = SysTimer();

	// Wait for handshake flag or timeout, whichever occurs first.
	do
	{
		status = IN_PORT_BYTE( IoBase + HOST_STATUS );
		if ( ( ( status & mask ) ^ xormask ) != 0 )
			break;
	}
	while ( ElapsedTime( StartTime ) < maxtime );

	// Max time elapsed, so we check handshake flag one final time in case this function
	// was interrupted for a long time.
	status = IN_PORT_BYTE( IoBase + HOST_STATUS );
	return ( ( ( status & mask ) ^ xormask ) != 0 );
}

////////////////////////////////////////////////////////////////////////////////////
// Handshake command string from CmdQ[] onto board.

void ClsPci::HandshakeCmdString()
{
	int		i;
	long	CmdBufAdrs;								// SmartAD RAM buffer address.
	char	*pQ;									// Ptr to command string.
	short	nCmdBytes = (short)( pCmdQ - CmdQ );	// Command string length.

	// If no CRMT fault was previously detected and command string is not empty ...
	if ( ( ( MyFaultFlags & FAULT_CRMT ) == 0 ) && ( nCmdBytes != 0 ) )
	{
		// Wait for CRMT or timeout, whichever occurs first.  If CRMT is asserted ...
		if ( HandshakeWait( (const char)STAT_CRMT, 0, MAXHSTIME ) )
		{
			// Copy command string from CmdQ[] to Smart A/D dual-port RAM command buffer.
			CmdBufAdrs	= MemBase + CMDBUF;			// Init RAM buffer address.
			pQ = CmdQ;								// Init ptr to command string.
			for ( i = 0; i < nCmdBytes; i++ )		// Copy command string.
			{
				OUT_PORT_BYTE( CmdBufAdrs++, *pQ++ );
			}

			// Negate CRMT handshake flag to release command to Smart A/D board for execution.
			OUT_PORT_BYTE( MemBase + MEM_HANDSHAKE, 0 );

			// Prepare for next command string.
			pCmdQ = CmdQ;
		}
		else
		{
			// Timeout: declare CRMT fault and terminate command handshake.
			DeclareError( FAULT_CRMT );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////
// Execute pending command string and fetch command response.

void ClsPci::HandshakeDatString( short nBytes )
{
	int		i;
	long	pSmadBuf;

	// If no DAV fault was previously detected and non-empty response string is expected ...
	if ( ( ( MyFaultFlags & FAULT_DAV ) == 0 ) && ( nBytes != 0 ) )
	{
		// Wait for DAV or timeout, whichever occurs first.  If DAV is asserted ...
		if ( HandshakeWait( STAT_DAV, 0, MAXHSTIME ) )
		{
			// Fetch all expected response bytes from Smart A/D board and store in DatQ[] starting
			// at NextWrite location.  Leave NextWrite ptr set to next free location in DatQ[].
			pSmadBuf = MemBase;
			for ( i = 0; i < nBytes; i++ )
			{
				*pDatQ++ = IN_PORT_BYTE( pSmadBuf++ );
			}

			// Negate DAV handshake flag on Smart A/D board.
			IN_PORT_BYTE( MemBase + MEM_HANDSHAKE );





//			double StartTime = SysTimer();
//			while ( ElapsedTime( StartTime ) < 0.001 );




		}
		else
		{
			// Timeout: declare DAV fault and terminate response handshake.
			DeclareError( FAULT_DAV );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////
// Execute pending command string in CmdQ[] and fetch command response into DatQ[].

void ClsPci::DeclareResponse( short nBytes )
{
	HandshakeCmdString();
	HandshakeDatString( nBytes );
}

//////////////////////////////////////////////////////////////////////////////
// Retrieve command response string into DatQ[].

void ClsPci::CmdClose()
{
	HandshakeCmdString();
	pDatQ = DatQ;
}
