//////////////////////////////////////////////////////////////////////////////
// Module	: ClsSmad.h														//
// Function	: Base class for all Sensoray Smart A/D boards.					//
// Author	: Jim Lamberson													//
// Created	: 08/03/1998													//
//////////////////////////////////////////////////////////////////////////////

#ifndef DEF_ClsSmad
#define DEF_ClsSmad

#include "SmartADDefs.h"			// Header for Smart A/D interface class.
#include "SmadEnums.h"				// Enumerated types.

// Configuration constants ----------------------------------------------------------------

#define MAX_NUMCHANS		32		// Maximum allowed channel count on any Smart A/D product.
#define DATBUF_SIZE			1500	// Size of response buffer; large enough to hold Ethernet packet.
#define CMDBUF_SIZE			32		// Size of command buffer.
#define MAXHSTIME			1.0		// Max time, in seconds, to wait for DAV or CRMT handshake.

// Standardized Logical Opcodes -----------------------------------------------------------

#define CMD_UNSUPPORTED		-1		// Unsupported command
#define CMD_RDCHAN			 0		// Get data from 1 channel
#define CMD_DFNTYP			 1		// Declare sensor type
#define CMD_DFNLIM			 2		// Set alarm limits
#define CMD_RDALRM			 3		// Get alarm flags for 8 channels
#define CMD_RDTEMP			 4		// Get TB temperature(s)
#define CMD_DFNOPN			 5		// Set open-sensor values
#define CMD_TARE			 6		// Tare gage
#define CMD_FILT50			 7		// Set 50Hz filter mode
#define CMD_READ8			 8		// Get data from 8 channels
#define CMD_CAL				10		// Calibration function
#define CMD_DIAGNOSE		11		// Diagnostic function
#define CMD_RDSLOP			12		// Get gage slope/intercept
#define CMD_WTSLOP			13		// Set gage slope/intercept
#define CMD_GZERO			14		// Set gage zero
#define CMD_GSPAN			15		// Set gage span
#define CMD_STFILT			16		// Set filter factor
#define CMD_SLEEP			17		// Switch to low-power standby
#define CMD_WAKEUP			18		// Wake from low-power standby
#define CMD_WCOEFF			19		// Set custom-R coefficients

// Diagnostics sub-opcodes ----------------------------------------------------------------

#define DIAG_GETRAWDATA		1		// Return channel's raw data.
#define DIAG_GETPRODUCTID	4		// Return product model number.
#define DIAG_GETVERSION		5		// Return firmware version number.
#define DIAG_HISPEEDMODE	8		// Switch to high-speed mode.

// Types ----------------------------------------------------------------------------------

typedef struct {				// CHANNEL ATTRIBUTES
	short	LogDefn;			//  Logical sensor definition code.
	short	HiLim;				//  High alarm limit.
	short	LoLim;				//  Low alarm limit.
	bool	LoLimEnable;		//  High alarm limit enabled.
	bool	HiLimEnable;		//  Low alarm limit enabled.
	bool	LoAlarm;			//  High alarm detected.
	bool	HiAlarm;			//  Low alarm detected.
	bool	FailHigh;			//  Fail high (versus low) on open sensor detected.
	char	Filter;				//  Filter factor.
	short	GagemVperV;			//	Gage mV/V * 10 (STDbus products only).
	short	GageMaxLoad;		//	Data value expected at max gage load (STDbus only).
} CHSTRUCT;

////////////////////////////////////////////////////////////////////////////////////////////
// SMART A/D BASE CLASS.
// This is the abstract base class for all Smart A/D boards.  Virtual command functions
// are executed by calling functions that are defined in derived classes.  In cases where
// derived class functions return data from a board, the data is normalized (converted to
// a uniform response string format) and placed in DatQ[] by the derived functions.  The
// base class then parses the response string using functions such as ReadByte(),
// ReadWord() and ReadReal().
////////////////////////////////////////////////////////////////////////////////////////////

class ClsSmad {
public:
	ClsSmad(short ProductID, long Adrs, short NumCh, const char *pSdcTbl, const char *pCmdTbl, char NumBoards = 1);
	virtual ~ClsSmad()	{}	// Destructor is virtual so that all derived class destructors will execute.
	double				AmbientTempGet		(short ChGroup);
	long				BoardAdrs			();
	double				Calibrate			(short ch, char range, double val);
	virtual void		CoefficientsSet		(short ch, double a, double b, double c)	{}
	void				Diagnostic			(short ch, char DiagID, char Parm, char RespSize, char *pBuf);
	bool				FailModeGet			(short ch);
	void				FailModeSet			(short ch, bool failHi);
	long				FaultFlagsGet		();
	char				FilterGet			(short ch);
	void				FilterSet			(short ch, char Val);
	virtual void		GageCalGet			(short ch, double *slope, short *offset)	{ *slope = 0.0; *offset = 0; }	// Non-STD.
	virtual void		GageCalSet			(short ch, double slope, short offset)		{}								// Non-STD.
	virtual void		GageSpanSet			(short ch, short val)						{}
	virtual void		GageTareSet			(short ch)									{}
	virtual void		GageZeroSet			(short ch)									{}
	bool				HiSpeedModeGet		();
	void				HiSpeedModeSet		();
	short				LimitAlarmsGet		(short *AlarmList);
	bool				LimitHiEnabGet		(short ch);
	void				LimitHiEnabSet		(short ch, bool Enable);
	bool				LimitLoEnabGet		(short ch);
	void				LimitLoEnabSet		(short ch, bool Enable);
	void				LimitValuesSet		(short ch, double LoVal, double HiVal);
	void				LoPwrStandbySet		(bool Standby);
	virtual bool		NetResetChk			()											{ return false; }	// Net only.
	virtual long		NetTimeoutGet		()											{ return 0; }		// Net only.
	virtual void		NetTimeoutSet		(long msec)									{}					// Net only.
	virtual short		NetVersionGet		()											{ return 0; }		// Net only.
	virtual short		ProductIDGet		()											{ return 0; }		// Product specific.
	void				RawDataAllGet		(short *pData);
	bool				Reject50HzGet		();
	void				Reject50HzSet		();
	void				ResetBoard			();
	void				SensorsAllGet		(double *pData);
	double				SensorGet			(short ch);
	short				SensorTypeGet		(short ch);
	virtual void		SensorTypeSet		(short ch, short logSDC)					{}
	void				SetErrCallback		(FPTR pErrHandler);
	short				StatusFlagsGet		();
	short				VersionGet			();
	// These public diagnostics are intended for use only by internal Sensoray test programs.
	void				GetAllRawData		();
	double				Linearize			(short LogSDC, short RawDat);
	short				RawChanGet			(short ch);
	virtual char		InByteDirect		(long adrs)									{ return 0; }
	virtual void		OutByteDirect		(long adrs, char val)						{}
protected:
	// Helper functions that vary for different types of Smart A/Ds:
	virtual	bool		ActiveALARM			()											{ return 0; }
	virtual void		CmdOpen				()											{ pCmdQ = CmdQ; pDatQ = DatQ; }
	virtual void		CmdClose			()											{}
	virtual void		CmdResetBoard		()											{}
	virtual	unsigned char CmdStatusFlagsGet	()											{ return 0; }
	virtual bool		HandshakeWait		(const char mask, const char xormask, const double maxtime)	{ return 0; }	// Non-net only.
	virtual void		DeclareResponse		(short nResponseBytes)						{}
	virtual void		SelectBoard			(short Board)								{}		// Net only.
	virtual void		SelectBoardByChan	(short *ch)									{}		// Net only.
	virtual void		SelectBoardByGroup	(short *Group)								{}		// Net only.
	virtual void		SendByte			(char val)									{}
	// Functions that are identical for all Smart A/Ds:
	void				DeclareError		(long ErrFlags);
	double				ElapsedTime			(double StartTime);
	void				InitChanProperties	(short ch, short LogSDC);
	bool				IsValidChan			(short Chan);
	short				LimTo32K			(double Value);
	char				ReadByte			()											{ return *pDatQ++; }
	short				ReadWord			();
	short				ReadWordRev			();
	double				ReadReal			();
	void				ResetBoardProperties();
	void				SendReal			(double N);
	// Functions implemented in derived, 0S-dependent class.
	virtual void		SysSleep			(long msec)									= 0;
	virtual double		SysTimer			()											= 0;
	// Overloaded command handshake functions:
	void				SendWord			(short val);
	void				SendWordRev			(short val);
	// Storage common to all Smart A/D boards:
	long				MyAddress;			// PCI bus & slot, IP adrs, or BasePort of Smart A/D board.
	short				MyProductID;		// Product model number as declared by object creator.
	long				MyFaultFlags;		// Board fault flags.
	short				MyBoardState;		// Board state flags (i.e., Reject50Hz, LoPwrStandby).
	long				MyNetTimeout;		// Network timeout interval, in milliseconds.
	char				MyNumBoards;		// Number of sub-boards (i.e., Model 2519 has two 518 sub boards).
	short				MyNumChans;			// Number of sensor channels on this product.
	short				MyNumGroups;		// Number of 8-channel groups.
	const char			*PhysSDC;			// Ptr to logical-to-physical Sensor Definition Code mapping table.
	const char			*Cmd;				// Ptr to logical-to-physical command opcode mapping table.
	CHSTRUCT			Ch[MAX_NUMCHANS];	// Sensor channel attributes.
	char				CmdQ[CMDBUF_SIZE];	// Command bytes to be handshaked to board.
	char				*pCmdQ;				// Ptr to next-write location in CmdQ[].
	char				DatQ[DATBUF_SIZE];	// Command response bytes that have been handshaked from board.
	char				*pDatQ;				// Ptr to next-write location in DatQ[].
	FPTR				pMyErrHandler;		// Pointer to external error handler (set to zero if no handler available).
private:
	// Storage:
	bool				MyIsModeHighSpeed;	// HighSpeed mode is active.
	bool				MyIsModeReject50Hz;	// LowSpeed mode's integration time set to 20ms (vs. default 16.7ms).
};

#endif // #ifndef DEF_ClsSmad
